/*
 * Copyright (c) 1990 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include "dixie.h"
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <quipu/commonarg.h>
#include <quipu/attrvalue.h>
#include <quipu/ds_error.h>
#include <quipu/bind.h>
#include <quipu/compare.h>

extern int	stream_connected;
extern int	debug;
extern int	bound;

#define BIND_END(x)	{ dn_free(thedn); return(x); }

struct ds_bind_arg	bind_arg;
struct ds_bind_arg	bind_result;
struct ds_bind_error	bind_error;

DN	bound_dn;
char	*bound_dnstr;
char	*bound_pw;

/*
 * bindop -- fork a child and bind it to the directory.  The child
 * is forked, allocates a stream connection to be used for future
 * communication, binds to the directory, and arranges to have the proper
 * acknowledgement sent back to the client.  This is either an error
 * or the ip address and port number to call back on.
 */

bindop(bp, rp, s)
char	*bp, *rp;
int	*s;
{
	struct sockaddr_in	addr;
	int			addrlen;
	int			s2, pid;
	DN			thedn = NULLDN;
	struct hostent		*hp;
	struct in_addr		myaddr;
	char			myname[32];
	char			*tmpdnstr;

	*rp = '\0';

	/* unbind first -- parent will rebind later */
	ds_unbind();

	switch ( pid = fork() ) {
	/* child -- successful fork */
	case  0:
		if ( debug ) printf("child -- forked ok\n");
		/* get a stream socket to communicate with */
		if ( (s2 = socket(AF_INET, SOCK_STREAM, 0)) == -1 )
			BIND_END(DXI_ERROR);

		addr.sin_family = AF_INET;
		addr.sin_addr.s_addr = INADDR_ANY;
		addr.sin_port = 0;

		addrlen = sizeof(addr);
		if ( bind(s2, &addr, addrlen) != 0 )
			BIND_END(DXI_ERROR);
		if ( getsockname(s2, &addr, &addrlen) != 0 )
			BIND_END(DXI_ERROR);

		/* construct x500 bind args */
		bind_arg.dba_version = DBA_VERSION_V1988;
		bind_arg.dba_auth_type = DBA_AUTH_SIMPLE;
		
		/* the dn to bind as */
		if ( *bp == '@' ) bp++;
		if ( (thedn = str2dn(bp)) == NULLDN && *bp != '\0' ) {
			if ( debug > 1 ) printf("bindop: str2dn(%s) failed\n",bp);
			BIND_END(DXI_BADDN);
		}
		bind_arg.dba_dn = thedn;
		tmpdnstr = strdup( bp );

		bp += strlen(bp) + 1;

		/* the password */
		strncpy(bind_arg.dba_passwd, bp, DBA_MAX_PASSWD_LEN);
		bind_arg.dba_passwd_len = strlen(bp);
		if (bind_arg.dba_passwd_len == 0)
			bind_arg.dba_auth_type = DBA_AUTH_NONE;

		/* attempt to bind */
		if ( debug ) printf("going for the bind...\n");
		if ( secure_ds_bind(&bind_arg, &bind_error, &bind_result) 
		    != DS_OK ) {
			if ( debug ) printf("bindop: bind failed\n");
			if (bind_error.dbe_type == DBE_TYPE_SECURITY)
				BIND_END(DXI_REFUSED)
			else
				BIND_END(DXI_DSA_DOWN)
		}
		if ( debug ) printf("bind successful!\n");
		/* record the dn we bound as */
		bound_dn = dn_cpy(thedn);
		bound_dnstr = tmpdnstr;
		if ( bound_pw )
			free( bound_pw );
		bound_pw = strdup( bp );

		gethostname(myname, sizeof(myname));
		if ( (hp = gethostbyname(myname)) == NULL ) {
			if ( debug ) perror("gethostname");
			clean_exit(1);
		}
		bcopy(hp->h_addr, (char *) &myaddr, sizeof(struct in_addr));

		/* we're bound -- arrange to send back the ack */
		/* XXX */
		sprintf(rp, "%s%c%d", inet_ntoa(myaddr), DXI_VAL_SEP, 
		    htons(addr.sin_port));
		if ( debug ) printf("bindop: rp is (%s)\n", rp);
		if ( debug ) printf("bindop: port is (%d)\n", addr.sin_port);

		*s = s2;
		BIND_END(DXI_SUCCESS);
		break;

	/* parent -- fork failed */
	case -1:
		nullbind();
		return(-1);
		break;

	/* parent -- fork succeeded */
	default:
		/* rebind parent as nulldn */
		if ( debug ) printf("forked child with pid %d\n", pid);
		nullbind();
		return(0);
		break;
	}
	/* NOT REACHED */
	return(-1);
}


/*
 * bbindop -- bind to a different directory entry when already bound.
 * There's no need to fork a child in this case, we just call the directory
 * bind routine with new parameters, first unbinding from our current
 * association.
 */

bbindop(bp, rp, check)
char	*bp, *rp;
int	check;
{
	DN	thedn = NULLDN;
	char	*tmpdnstr;

	*rp = '\0';
	if ( *bp == '@' ) bp++;

	if ( (thedn = str2dn(bp)) == NULLDN && *bp != '\0' ) {
		if ( debug > 1 ) printf("bindop: str2dn(%s) failed\n",bp);
		return(DXI_BADDN);
	}
	if ( check && dn_cmp(thedn, bound_dn) == OK )
		BIND_END(DXI_SUCCESS);
	tmpdnstr = strdup( bp );

	/* construct x500 bind args */
	bind_arg.dba_version = DBA_VERSION_V1988;
	bind_arg.dba_auth_type = DBA_AUTH_SIMPLE;
	bind_arg.dba_dn = thedn;

	bp += strlen(bp) + 1;

	/* the password */
	strncpy(bind_arg.dba_passwd, bp, DBA_MAX_PASSWD_LEN);
	bind_arg.dba_passwd_len = strlen(bp);
	if (bind_arg.dba_passwd_len == 0)
		bind_arg.dba_auth_type = DBA_AUTH_NONE;

	/* unbind first */
	ds_unbind();

	/* attempt to re-bind */
	if ( debug ) printf("bbinding...\n");
	if ( secure_ds_bind(&bind_arg, &bind_error, &bind_result) != DS_OK ) {
		if ( debug ) printf("bbindop: bind failed\n");
		if (bind_error.dbe_type == DBE_TYPE_SECURITY)
			BIND_END(DXI_REFUSED)
		else
			BIND_END(DXI_DSA_DOWN)
	}
	if ( debug ) printf("bbind successful!\n");
	if ( bound_dn != NULLDN )
		dn_free( bound_dn );
	bound_dn = dn_cpy(thedn);
	if ( bound_dnstr != NULL )
		free( bound_dnstr );
	bound_dnstr = tmpdnstr;
	if ( bound_pw )
		free( bound_pw );
	bound_pw = strdup( bp );

	BIND_END(DXI_SUCCESS);
}

/*
 * pwcompare -- compare the password to that of the entry specified in
 * bp.  Return 0 if they match, a nonzero indication of the error
 * otherwise.
 */

pwcompare( bp )
char	*bp;
{
	DN				thedn = NULLDN;
	struct ds_compare_arg		cma;
	struct ds_compare_result	cmr;
	struct DSError			cme;
	CommonArgs			*ca = &cma.cma_common;
	ServiceControl			*sc = &ca->ca_servicecontrol;
	AttributeType			pwtype;
	AttributeValue			pwvalue;
	int				i;

	if ( *bp == '@' ) bp++;

	if ( (thedn = str2dn(bp)) == NULLDN ) {
		if ( debug > 1 ) printf("bindop: str2dn(%s) failed\n",bp);
		return(DXI_BADDN);
	}
	cma.cma_object = thedn;

	bp += strlen( bp ) + 1;
	if ( *bp == '\0' )
		return( 0 );

	ca->ca_requestor = NULLDN;
	ca->ca_aliased_rdns = 0;
	ca->ca_security = ( struct security_parms * ) 0;
	ca->ca_sig = ( struct signature * ) 0;
	ca->ca_extensions = ( struct extension * ) 0;

	sc->svc_options = SVC_OPT_PREFERCHAIN;
	sc->svc_prio = SVC_PRIO_MED;
	sc->svc_timelimit = SVC_NOTIMELIMIT;
	sc->svc_sizelimit = SVC_NOSIZELIMIT;
	sc->svc_scopeofreferral = SVC_REFSCOPE_NONE;

	if ( (pwtype = str2AttrT( "userPassword" )) == NULLAttrT ) {
		if (debug) printf("str2AttrT of userPassword failed\n");
		dn_free( thedn );
		return( DXI_ERROR );
	}
	cma.cma_purported.ava_type = pwtype;
	if ( (pwvalue = str2AttrV( bp, pwtype->oa_syntax )) 
	    == NULLAttrV ) {
		if (debug) printf("str2AttrV of userPassword failed\n");
		dn_free( thedn );
		return( DXI_BAD_ATTRVAL );
	}
	cma.cma_purported.ava_value = pwvalue;

	for ( i = 0; i < 2; i++ ) {
		if ( ds_compare( &cma, &cme, &cmr ) == DS_OK )
			break;
		if ( debug & 2 ) printf("ds_compare failed\n");

		if ( cme.dse_type != DSE_REMOTEERROR || stream_connected )
			break;

		(void) nullbind();

		if ( bound == 0 )
			break;
	}
	dn_free( thedn );

	if ( debug & 2 )
		printf("compare successful - checking for errors\n");

	switch ( cme.dse_type ) {
	case DSE_REMOTEERROR:
		if ( debug & 2 ) {
			printf("\tservice error\n");
			error( &cme );
		}
		return(DXI_DSA_DOWN);
	case DSE_REFERRAL:
		if ( debug & 2 ) {
			printf("\treferral error\n");
			error( &cme );
		}
		return(DXI_REFER_ERROR);
	case DSE_NAMEERROR:
		if ( debug & 2 ) {
			printf("\tname error\n");
			error( &cme );
		}
		return(DXI_NAME_ERROR);
	case DSE_SERVICEERROR:
		if ( debug & 2 ) {
			printf("\tservice error\n");
			error( &cme );
		}
		break;
	case DSE_ATTRIBUTEERROR:
		if ( debug & 2 ) {
			printf("\tattribute error\n");
			error( &cme );
		}
		return(DXI_NEX_ATTR);
	case DSE_NOERROR:
		if ( debug & 2 ) printf("\tno error\n");
		break;
	default:
		if ( debug & 2 ) {
			printf("\tsome unknown error (%d)\n");
			error( &cme );
		}
		return(DXI_ERROR);
	}

	return( cmr.cmr_matched ? 0 : DXI_REFUSED );
}
