/*
 * Copyright (c) 1990 Regents of the University of Michigan.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that this notice is preserved and that due credit is given
 * to the University of Michigan at Ann Arbor. The name of the University
 * may not be used to endorse or promote products derived from this
 * software without specific prior written permission. This software
 * is provided ``as is'' without express or implied warranty.
 */

#include "dixie.h"
#include <stdio.h>
#include <sys/syslog.h>
#include <sys/time.h>
#include <quipu/ds_error.h>
#include <quipu/attrvalue.h>
#include <quipu/name.h>
#include <quipu/commonarg.h>
#include <quipu/bind.h>
#include <quipu/modify.h>
#include <quipu/read.h>
#include <ctype.h>

common_args	default_common;

union val {
	struct mailbox		*mailbox;
	struct fax		*fax;
	struct postaddr		*postaddr;
	struct telex		*telex;
	struct teletex		*teletex;
	struct pref_deliv	*pref_deliv;
	struct Guide		*guide;
};

static int	acl_syntax;
static int	objectclass_syntax;
static int	postaladdress_syntax;
static int	dn_syntax;
static int	mail_syntax;
static int	photo_syntax;
static int	jpegphoto_syntax;
static int	multilinedescription_syntax;
static int	audio_syntax;

static AttributeType 	objectclass_type;
int			want_objectclass;

extern int	dosyslog;
extern int	debug;
extern int	stream_connected;
extern long	stream_conn_start_time;

advise(s)
char	*s;
{
	printf( "Be advised:\n" );
	printf( "\t%s\n", s );
}

adios(s1, s2)
char	*s1, *s2;
{
	printf("Adios:\n");
	printf("%s : %s", s1, s2);
	clean_exit(1);
}


int
get_one_syntax( attrib, required )
    char *attrib;
    int  required;
{
	oid_table_attr	*p, *name2attr();

	if (( p = name2attr( attrib )) != (oid_table_attr *) 0)
	    return( p->oa_syntax );

	if ( !required ) {
	    return( -1 );
	}

	if ( debug ) printf("name2attr( \"%s\" ) failed - exiting\n", attrib );
	if ( dosyslog ) syslog(LOG_INFO|LOG_LOCAL1,
		"name2attr( \"%s\" ) failed - exiting", attrib );
	clean_exit( 1 );
}


get_syntaxes()
{
	oid_table_attr	*p, *name2attr();

	if (( objectclass_type = str2AttrT( "objectclass" )) == NULL ) {
	    if ( debug ) printf( "str2AttrT( \"objectclass\" failed - exiting\n" );
	    if ( dosyslog ) syslog( LOG_INFO|LOG_LOCAL1,
		    "str2AttrT( \"objectclass\" failed - exiting" );
	}
	
	acl_syntax = get_one_syntax( "acl", 1 );
	objectclass_syntax = get_one_syntax( "objectclass", 1 );
	postaladdress_syntax = get_one_syntax( "postaladdress", 0 );
	dn_syntax = get_one_syntax( "aliasedobjectname", 1 );
	mail_syntax = get_one_syntax( "mail", 0 );
	photo_syntax = get_one_syntax( "photo", 0 );
	jpegphoto_syntax = get_one_syntax( "jpegPhoto", 0 );
	audio_syntax = get_one_syntax( "audio", 0 );
	multilinedescription_syntax =
		get_one_syntax( "multiLineDescription", 0 );
}

/*
 * nullbind -- bind to the directory as NULLDN.  This routine is called
 * to bind initially, and if the connection is lost for any reason.  It
 * sets the global variable bound indicating success or failure to bind.
 */

nullbind()
{
	extern int		bound;
        struct ds_bind_arg      bind_arg;
        struct ds_bind_arg      bind_result;
        struct ds_bind_error    bind_error;

        /* construct the bind argument to bind as NULLDN */
        bind_arg.dba_version = DBA_VERSION_V1988;
        bind_arg.dba_auth_type = DBA_AUTH_NONE;
        bind_arg.dba_dn = NULLDN;

        /* bind to the directory -- may take a while if quipu is loading */
	if ( debug ) printf("(re)binding...\n");
	if ( dosyslog ) syslog( LOG_INFO|LOG_LOCAL1, "(re)binding" );
	if ( secure_ds_bind(&bind_arg, &bind_result, &bind_error) == DS_OK )
		bound = 1;
	else
		bound = 0;

	if ( dosyslog ) syslog( LOG_INFO|LOG_LOCAL1, "Bind was %s successful",
		( bound == 1 ? "" : "NOT" ));

        if ( debug ) {
                if ( bound == 1 )
			printf("Bind was successful\n");
		else
			printf("Bind was NOT successful\n");
        }
}

/*
 * entry_sprint -- step through the list of entries given by ep and print
 * the associated attribute types and values in printable form into the
 * buffer pointed to by bp.  The entries should have already been decoded.
 * A count of the number of entries is returned, -1 in case of some error.
 */

entry_sprint(bp, ep, bplenp, include_fileattrs, version, toobigp)
char		*bp;
EntryInfo	*ep;
int		*bplenp;
char		include_fileattrs;
int		version;
int		*toobigp;
{
	int	matches = 0;
	int	len, bplen;
	PS	ps;

	bplen = *bplenp;
	if ((ps = ps_alloc(str_open)) == NULLPS) {
		if (debug) printf("ps_alloc failed\n");
		clean_exit(1);
	}
	if (str_setup(ps, bp, bplen, 1) == NOTOK) {
		if (debug) printf("str_setup failed\n");
		clean_exit(1);
	}

	while ( ep ) {
		matches++;
		ps_printf(ps, "%c", DXI_DNS_SEP);
		bplen--;
		bp++;

		/* print the dn of this entry */
		len = dn_sprint(ps, ep->ent_dn);
		bplen -= len;
		bp += len;

		ps_printf(ps, "%c", DXI_ATR_SEP);
		bplen--;
		bp++;

		/* print the attr types and vals found */
		len = as_sprint(ps, bp, ep->ent_attr, bplen,
			include_fileattrs, version, toobigp);
		bplen -= len;
		bp += len;

		ep = ep->ent_next;
	}
	*ps->ps_ptr = '\0';

	ps_free(ps);
	*bplenp = *bplenp - bplen;
	return(matches);
}

/*
 * as_sprint -- copy the attributes and values specified in as in printed
 * form into the buffer pointed to by bp.  as should have already been 
 * decoded.  Values are written to bp like this:
 *	type1^Avalue1^A...^AvalueN^Btype2^Avalue1^A...^AvalueN\0
 * It is assumed that all values have simple string representations for
 * their decoded types.
 */

as_sprint(ps, bp, as, bplen, include_fileattrs, version, toobigp)
PS		ps;
char		*bp;
Attr_Sequence	as;
int		bplen;
char		include_fileattrs;
int		version;
int		*toobigp;
{
	AV_Sequence	av;
	char		*savebp, *save;
	char		*save_psptr, *save_psbase;
	int		len;
	char		*temp_bits, *atend_bits, *p;
	int		temp_len, atend_len, atend_count;
	char		photo_flags;
	int		along;
	long		photo_decode();
	extern IFP	oc_avsprint;

	*toobigp = 0;
	atend_count = atend_len = 0;
	atend_bits = NULL;

	if ( as == NULLATTR ) {
		*bp = '\0';
		return( 0 );
	}

	savebp = bp;
	save_psbase = ps->ps_ptr;
	while ( as ) {
		int	syntax;

		syntax = as->attr_type->oa_syntax;
		/* ignore acl */
/*
why are we ignoring acls?
		if ( syntax == acl_syntax || (syntax == objectclass_syntax
*/
		if ( syntax == objectclass_syntax && want_objectclass == 0 ) {
			as = as->attr_link;
			continue;
		}

		if ( (syntax == photo_syntax || syntax == jpegphoto_syntax
		    || syntax == audio_syntax) && include_fileattrs == 0 ) {
			as = as->attr_link;
			continue;
		}

		save_psptr = ps->ps_ptr;
		AttrT_print(ps, as->attr_type, EDBOUT);
		*ps->ps_ptr = '\0';

		if ( debug & 8 ) {
			printf("\t(%s) (%d)\n",bp, syntax);
			fflush(stdout);
		}
		bp += strlen(bp);

		if ( version < 1 && syntax == objectclass_syntax ) {
			ps_print(ps, "\001");
			save = ++bp;

			(*oc_avsprint)(ps, as->attr_value, READOUT);
			*ps->ps_ptr = '\0';

			if ( debug & 8 ) {
				printf("\t\t(%s)\n",save);
				printf("\t%d more bytes\n",
				    ps->ps_ptr - save_psptr);
				fflush(stdout);
			}

			bp += strlen(bp);
			ps_print(ps, "\002");

			as = as->attr_link;
			continue;
		}

		av = as->attr_value;
		while ( av ) {
			ps_print(ps, "\001");
			save = ++bp;

			/* distinguished name */
			if ( syntax == dn_syntax ) {
				dn_print(ps, (DN) av->avseq_av.av_struct,
				    EDBOUT);

			/* postal address */
			} else if ( syntax == postaladdress_syntax ) {
				AttrV_print(ps, av, EDBOUT);
				*ps->ps_ptr = '\0';

			/* mailbox */
			} else if ( syntax == mail_syntax ) {
				AttrV_print(ps, av, READOUT);

			/* multilinedescription */
			} else if ( syntax == multilinedescription_syntax ) {
				AttrV_print(ps, av, EDBOUT);

			/* photo */
			} else if ( syntax == photo_syntax ||
			    syntax == jpegphoto_syntax ||
			    syntax == audio_syntax ) {
				if ( debug & 8 )
				    printf( "decoding file attribute\n" );
				if ( syntax == audio_syntax )
				    temp_len = audio_decode( av, &temp_bits );
				else
				    temp_len = photo_decode(av, &temp_bits,
					&photo_flags );
				if ( temp_len > 0 ) {
				    if ( debug & 8 )
					printf( "decode ok\n" );
				    ps_print(ps, "\004" );
				    if (debug & 8 )
					if ( syntax != audio_syntax )
					    printf( "append photo, len=%d %s%s\n",
					    temp_len, ( photo_flags &
					    DXI_PHOTO_JPEG ) ? "JPEG" : "G3Fax",
					    ( photo_flags &
					    DXI_PHOTO_G3FAX_TWODIM) ?
					    "TwoDim" : "" );
					else
					    printf( "append audio, len=%d\n",
						temp_len );

				    temp_len += ( syntax == audio_syntax ) ?
					sizeof( long ) : 1 + sizeof( long );

				    if ( atend_bits == NULL ) {
					atend_len = temp_len;
					if (( atend_bits = malloc( atend_len )) == NULL ) {
					    if (debug) printf( "malloc fail" );
					    clean_exit( 1 );
					}
					p = atend_bits;
				    } else {
					if (( atend_bits = realloc( atend_bits,
						atend_len + temp_len))
						== NULL ) {
					    if (debug) printf( "realloc fail" );
					    clean_exit( 1 );
					}
					p = atend_bits + atend_len;
					atend_len += temp_len;
				    }

				    ++atend_count;
				    along = htonl( temp_len - sizeof( long ));
				    bcopy( &along, p, sizeof( along ));
				    p += sizeof( long );
				    if ( syntax != audio_syntax ) {
					*p++ = photo_flags;
					--temp_len;
				    }
				    bcopy( temp_bits, p, temp_len -
					    sizeof( long ));
				    free( temp_bits );
				} else {
				    if ( debug & 8 )
					printf( "decode failed\n" );
				}
			} else {
				AttrV_print(ps, av, EDBOUT);
			}
			*ps->ps_ptr = '\0';

			if ( debug & 8 ) {
				printf("\t\t(%s)\n",save);
				fflush(stdout);
			}

			bp += strlen(bp);
			av = av->avseq_next;
		}

		if ( debug & 8 ) {
			printf("\t%d more bytes\n", ps->ps_ptr - save_psptr);
			fflush(stdout);
		}

		ps_print(ps, "\002");
		as = as->attr_link;
	}

	len = ps->ps_ptr - save_psbase;
	if (len > 0) {
		*--ps->ps_ptr = '\0';
		--len;
	} else {
		*ps->ps_ptr = '\0';
		len = 0;
	}

	if ( atend_len > 0 ) {
	    ++len;
	    p = savebp + len;

	    if ( atend_len > bplen ) {
		if (debug & 8 ) printf( "atend stuff too large\n" );
		*toobigp = 1;
		along = htonl( 0 );
		while ( atend_count-- > 0 ) {
		    bcopy( &along, p, sizeof( long ));
		    p += sizeof( long );
		    len += sizeof( long );
		}
	    } else {
		if (debug & 8 ) printf( "appending atend values\n" );
		bcopy( atend_bits, p, atend_len );
		len += atend_len;
	    }

	    free( atend_bits );
	}

	if (debug & 8) {
		printf("as_sprint of attributes produces %d bytes\n", len);
	}

	return(len);
}


/*
 * decode an audio syntax, set bitsp, return length
 */
long
audio_decode( av, bitsp )
    AttributeValue	av;
    char		**bitsp;
{
    PE		pe;
    struct qbuf	*qb, *p;
    long	len;

    qb = (struct qbuf *)
		(((struct file_syntax *)av->av_struct)->fs_attr->av_struct);

    len = 0;
    for (p = qb -> qb_forw; p != qb; p = p -> qb_forw) {
	len += p->qb_len;
    }

    if (( *bitsp = malloc( len )) == NULL ) {
	if ( debug & 8 )
	    printf( "malloc for audio data failed\n" );
	return( -1 );
    }

    len = 0;
    for (p = qb -> qb_forw; p != qb; p = p -> qb_forw) {
	bcopy( p->qb_data, *bitsp+len, p->qb_len );
	len += p->qb_len;
    }

    return( len );
}


/*
 * decode a photo, set bitsp , return length
 * flagsp are set to indicate:
 *   G3FAX (default):		0
 *   G3FAX two dimensional:	DXI_PHOTO_G3FAX_TWODIM
 *   JPEG format:		DXI_PHOTO_JPEG
 * (mostly lifted from isode/others/quipu/photo/decode.c)
*/
long
photo_decode ( av, bitsp, flagsp )
    AttributeValue	av;
    char		**bitsp;
    char		*flagsp;
{
    long	res;
    PE		pe, set, member, seq;
    int		twoDimensional;

    *flagsp = 0;
    pe = (PE)(((struct file_syntax *)av->av_struct)->fs_attr->av_struct);

    if ( debug & 8 ) {
	printf( "photo_decode  pe_class: %d  pe_form: %d  pe_id: %d\n",
	    pe->pe_class, pe->pe_form, pe->pe_id );
    }

    if (pe->pe_class == PE_CLASS_UNIV && pe->pe_form == PE_FORM_PRIM &&
	pe->pe_id == PE_PRIM_BITS) { /* old  BIT STRING-like form */
	if (pe_pullup (pe) == NOTOK) {
	    return( -1 );
        }
	*bitsp = (char *)pe->pe_prim;
	*flagsp = DXI_PHOTO_G3FAX_TWODIM;
	res = ps_get_abs(pe);

    } else if ( (pe->pe_class == PE_CLASS_UNIV || pe->pe_class == PE_CLASS_CONT)
	     && pe->pe_form == PE_FORM_PRIM &&
	     pe->pe_id == PE_PRIM_OCTS) { /* JPEG photo */
	if (pe_pullup (pe) == NOTOK) {
	    return( -1 );
        }
	*bitsp = (char *)pe->pe_prim;
	*flagsp = DXI_PHOTO_JPEG;
	res = ps_get_abs(pe);

    } else if (pe->pe_class == PE_CLASS_CONT && pe->pe_form == PE_FORM_CONS &&
	     pe->pe_id == 3) {

	/* first member must be SET which describes G3-Fax options */

	set = first_member (pe);
	if (!set || set->pe_class != PE_CLASS_UNIV ||
	    set->pe_form != PE_FORM_CONS || set->pe_id != PE_CONS_SET) {
	    return( -1 );
	}

	for (member = first_member (set); member; member = next_member (set, member)) {
	    if (member->pe_class == PE_CLASS_CONT &&
		member->pe_form == PE_FORM_PRIM && member->pe_id == 1) {
	        twoDimensional = bit_test (prim2bit (member), 8);
		if (twoDimensional == -1) {
		    return( -1 );
	        }
	    }
	}

	/* SEQUENCE of BIT STRING should follow SET */

	seq = next_member (pe, set);
	if (!seq || seq->pe_class != PE_CLASS_UNIV ||
	    seq->pe_form != PE_FORM_CONS || seq->pe_id != PE_CONS_SEQ) {
	    return( -1 );
	}

	for (member = first_member (seq); member; member = next_member (seq, member)) {
	    if (member->pe_class != PE_CLASS_UNIV ||
		member->pe_form != PE_FORM_PRIM ||
		member->pe_id != PE_PRIM_BITS) {
		return( -1 );
	    }
	    *bitsp = (char *)(member->pe_prim) + 1;
	    if ( twoDimensional ) {
		*flagsp = DXI_PHOTO_G3FAX_TWODIM;
	    }
	    res = ps_get_abs(member) - 1;
	}
    } else
	res = -1;

    return( res );
}

/*
 * dn_sprint -- copy the distinguished name specified in dn in printed
 * form to the buffer pointed to by bp.  dn should have already been
 * decoded.  The value written is terminated by a '\0'.
 */

dn_sprint(ps, dn)
PS	ps;
DN	dn;
{
	char	*save;
	int	len;

	save = ps->ps_ptr;
	dn_print(ps, dn, EDBOUT);
	*ps->ps_ptr = '\0';
	len = ps->ps_ptr - save;

	if (debug & 8)
		printf("dn_sprint returns %d bytes\n", len);
	return(len);
}

/*
 * as_typelist -- take a null separated list of attribute types and
 * return the corresponding Attr_Sequence list.  Bad attributes (ones
 * that cannot be converted to an AttrT) are ignored.  The list is
 * terminated by two nulls in a row.  A list with no attribute types
 * in it means that all attributes are wanted.  allflg is changed
 * appropriately.
 */

Attr_Sequence as_typelist(bp, allflg)
char	*bp;
int	*allflg;
{
	Attr_Sequence	types, tempas, as_merge(), as_comp_new();
	AttributeType	type;

	/* check if all attributes wanted */
	if ( bp == NULL || *bp == '\0' ) {
		*allflg = 1;
		want_objectclass = 1;
		return( NULLATTR );
	}

	want_objectclass = 0;
	*allflg = 0;
	/* get the first valid attribute */
	while ( (type = str2AttrT(bp)) == NULLAttrT ) {
		if ( debug )
			printf("str2AttrT of (%s) failed\n", bp);
		bp += strlen(bp) + 1;

		/* we reached the end without finding one */
		if ( *bp == '\0' ) {
			if ( debug )
				printf("No valid attributes\n");
			return(NULLATTR);
		}
	}
	if ( type == objectclass_type )
		want_objectclass = 1;
	types = as_comp_new(type, NULLAV, NULLACL_INFO);
	if ( types == NULLATTR ) {
		if ( debug )
			printf("as_comp_new failed\n");
		return(NULLATTR);
	}
	
	/* tack on the rest of the attributes */
	bp += strlen(bp) + 1;
	while ( *bp != '\0' ) {
		if ( (type = str2AttrT(bp)) == NULLAttrT ) {
			if ( debug )
				printf("str2AttrT(%s) failed\n", bp);
		} else {
			if ( type == objectclass_type )
				want_objectclass = 1;
			tempas = as_comp_new(type, NULLAV, NULLACL_INFO);
			if ( tempas == NULLATTR ) {
				if ( debug )
					printf("as_comp_new failed\n");
				return(NULLATTR);
			} else
				types = as_merge(types, tempas);
		}

		bp += strlen(bp) + 1;
	}

	return(types);
}

em_free(em)
struct entrymod	*em;
{
	struct entrymod	*next;

	while ( em ) {
		next = em->em_next;
		as_free(em->em_what);
		free(em);
		em = next;
	}
	return(0);
}

ei_free( ep )
EntryInfo	*ep;
{
	if ( ep == NULLENTRYINFO )
		return;

	dn_free( ep->ent_dn );
	as_free( ep->ent_attr );
}

/*
 * Print arbitrary stuff, for debugging.
 */

#define BPLEN	48

bprint( data, len )
    char	*data;
    int		len;
{
    static char	hexdig[] = "0123456789abcdef";
    char	out[ BPLEN ];
    int		i = 0;

    bzero( out, BPLEN );
    for ( ;; ) {
	if ( len < 1 ) {
	    printf( "%s\n", ( i == 0 ) ? "(end)" : out );
	    break;
	}

	if ( isgraph( (unsigned char)*data )) {
	    out[ i ] = ' ';
	    out[ i+1 ] = *data;
	} else {
	    out[ i ] = hexdig[ ( *data & 0xf0 ) >> 4 ];
	    out[ i+1 ] = hexdig[ *data & 0x0f ];
	}
	i += 2;
	len--;
	data++;

	if ( i > BPLEN - 2 ) {
	    printf( "%s\n", out );
	    bzero( out, BPLEN );
	    i = 0;
	    continue;
	}
	out[ i++ ] = ' ';
    }
}

/*
 * typematch - search the Attr_Sequence as for a type type with
 * value value.  Return 1 if found, 0 otherwise.
 */

typematch( ap, type, value )
Attr_Sequence	ap;
AttributeType	type;
AttributeValue	value;
{
	AV_Sequence	vp;

	if ( debug & 8 ) printf( "typematch\n" );

	while ( ap ) {
		if ( debug & 8 )
			printf( "matching (%s)\n", 
			    ap->attr_type->oa_ot.ot_name );
		if ( AttrT_cmp( ap->attr_type, type ) == 0 ) {
			if ( debug & 8 ) printf( "\tfound attr match\n" );
			vp = ap->attr_value;
			while ( vp ) {
				if ( AttrV_cmp( &vp->avseq_av, value ) == 0 ) {
					if ( debug & 8 )
						printf( "found val match\n" );
					return( 1 );
				}
				vp = vp->avseq_next;
			}
		}
		ap = ap->attr_link;
	}

	return( 0 );
}

/*
 * read_entry - read entry specified by dn and return attributes specified
 * by attr.  The result of the read, an entry info struct is returned,
 * or null if there was some problem.  If attr is null, all attributes
 * are returned.  If vals is nonzero, values are returned with attributes,
 * otherwise, just attribute types are returned.
 */
/* andy added last parameter*/

EntryInfo *read_entry( dn, attr, vals, all, dsaup, err, creferr,dontderef )
DN		dn;
Attr_Sequence	attr;
int		vals;
int		all;
int		*dsaup;
int		*err;
struct DSError	**creferr;
char 		dontderef;
{
	struct ds_read_arg	ra;
	struct ds_read_result	rr;
	EntryInfo		*ei;
	static struct DSError	re;
	CommonArgs		*ca = &ra.rda_common;
	ServiceControl		*sc = &ra.rda_common.ca_servicecontrol;

	if ( debug & 2 ) printf( "read_entry\n" );

	sc->svc_options = SVC_OPT_PREFERCHAIN;
	sc->svc_prio = SVC_PRIO_MED;
	sc->svc_timelimit = SVC_NOTIMELIMIT;
	sc->svc_sizelimit = SVC_NOSIZELIMIT;
	sc->svc_scopeofreferral = SVC_REFSCOPE_NONE;
	/*andy*/
	if (dontderef) {
	   sc->svc_options |= SVC_OPT_DONTDEREFERENCEALIAS;
	   if (debug) printf("DONTDEREF\n");
        }

	ca->ca_requestor = NULLDN;
	ca->ca_aliased_rdns = 0;
	ca->ca_security = ( struct security_parms * ) NULL;
	ca->ca_sig = ( struct signature * ) NULL;
	ca->ca_extensions = ( struct extension * ) NULL;

	ra.rda_object = dn;
	ra.rda_eis.eis_allattributes = all;
	ra.rda_eis.eis_select = attr;
	if ( vals )
		ra.rda_eis.eis_infotypes = EIS_ATTRIBUTESANDVALUES;
	else
		ra.rda_eis.eis_infotypes = EIS_ATTRIBUTETYPESONLY;

	*dsaup = 1;
	if ( ds_read( &ra, &re, &rr ) != DS_OK ) {
		if (re.dse_type == DSE_REMOTEERROR)
			*dsaup = 0;
		if ( debug )
			printf( "read_entry: ds_read failed\n" );
		switch ( re.dse_type ) {
		case DSE_REMOTEERROR:
			if ( debug & 2 ) {
				printf("\tservice error\n");
				error( &re );
			}
			*err = DXI_DSA_DOWN;
			break;
		case DSE_REFERRAL:
			if ( debug & 2 ) {
				printf("\treferral error\n");
			}
			*err = DXI_REFER_ERROR;
			if ( creferr != 0 )
				*creferr = &re;
			else
				error( &re );
			break;
		case DSE_NAMEERROR:
			if ( debug & 2 ) {
				printf("\tname error\n");
				error( &re );
			}
			*err = DXI_NAME_ERROR;
			break;
		case DSE_SERVICEERROR:
			if ( debug & 2 ) {
				printf("\tservice error\n");
				error( &re );
			}
			if ( re.dse_un.dse_un_service.DSE_sv_problem
			    == DSE_SV_TIMELIMITEXCEEDED )
				*err = DXI_TIMELIMIT;
			else if ( re.dse_un.dse_un_service.DSE_sv_problem
			    == DSE_SV_ADMINLIMITEXCEEDED )
				*err = DXI_SIZELIMIT;
			else
				*err = DXI_ERROR;
			break;
		case DSE_ATTRIBUTEERROR:
			if ( debug & 2 ) {
				printf("\tattribute error\n");
				error( &re );
			}
			*err = DXI_NEX_ATTR;
			break;
		case DSE_NOERROR:
			if ( debug & 2 ) printf("\tno error\n");
			*err = DXI_ERROR;
			break;
		default:
			if ( debug & 2 ) {
				printf("\tsome unknown error (%d)\n",
				    re.dse_type);
				error( &re );
			}
			*err = DXI_ERROR;
			break;
		}
		return( NULLENTRYINFO );
	}

	ei = entryinfo_alloc();
	bcopy( (char *) &rr.rdr_entry, (char *) ei, sizeof(EntryInfo) );
	ei->ent_next = NULLENTRYINFO;

	return( ei );
}

/*
 * dereference_aliases -- search through the list of entries pointed to
 * by ep and replace entries that are aliases by the entries that the
 * alias points to.
 */

/*
dereference_aliases( ep, as )
EntryInfo	**ep;
Attr_Sequence	as;
{
	EntryInfo		*cur, *prev, *deref;
	static AttributeType	octype;
	static AttributeValue	ocval;

	if ( debug & 8 ) printf( "deference_aliases\n" );

	if ( *ep == NULLENTRYINFO )
		return;

	if ( octype == NULLAttrT ) {
		if ( (octype = str2AttrT( "objectClass" )) == NULLAttrT ) {
			if ( debug ) printf( "str2AttrT failed\n" );
			clean_exit( 1 );
		}
		if ( (ocval = str2AttrV( "alias", 
		    octype->oa_syntax )) == NULLAttrV ) {
			if ( debug ) printf( "str2AttrV failed\n" );
			clean_exit( 1 );
		}
	}

	prev = NULLENTRYINFO;
	cur = *ep;
	while ( cur ) {
		if ( typematch( cur->ent_attr, octype, ocval ) == 0 ) {
			prev = cur;
			cur = cur->ent_next;
			continue;
		}

		if ( debug & 8 ) printf( "found an alias\n" );
		if ( (deref = read_entry( cur->ent_dn, as, 1 ))
		    != NULLENTRYINFO ) {
			if ( debug & 8 )
			    printf( "replacing alias entry\n" );
			deref->ent_next = cur->ent_next;
			if ( prev == NULLENTRYINFO )
				*ep = deref;
			else
				prev->ent_next = deref;
			ei_free( cur );
			cur = deref;
		}

		prev = cur;
		cur = cur->ent_next;
	}
}
*/

count_matches(ep, type, value, mp)
EntryInfo	*ep;		/* the entries to check */
AttributeType	type;
AttributeValue	value;
EntryInfo	**mp;		/* the list of matches to return */
{
	int		matches = 0;
	EntryInfo	*tmp, *savenext;
	Attr_Sequence	ap;
	AV_Sequence	vp;

	if ( debug & 8 )
		printf( "looking for (%s) with value (%s)\n",
		    type->oa_ot.ot_name, value->av_struct );

	*mp = NULLENTRYINFO;
	while (ep) {
		if ( debug & 8 ) printf("Entry\n");
		savenext = ep->ent_next;
		ap = ep->ent_attr;
		while (ap) {
			if ( debug & 8 )
				printf( "\tlooking at attr (%s)\n",
				    ap->attr_type->oa_ot.ot_name );
			if ( AttrT_cmp(type, ap->attr_type) == 0 ) 
				break;
			else 
				ap = ap->attr_link;
		}

		if (ap) {
			if ( debug & 8 ) printf("\tAttribute found\n");
			vp = ap->attr_value;

			while (vp) {
				if (debug & 8) printf("\t\tValue is (%s)\n",
				    vp->avseq_av.av_struct);
				if ( AttrV_cmp(value, &(vp->avseq_av)) == 0 ) {
					matches++;
					if ( debug & 8 ) printf("\t\t\tmatch is (%d)\n", matches);
					if ( *mp == NULLENTRYINFO ) {
						*mp = ep;
						(*mp)->ent_next = NULLENTRYINFO;
					} else {
						tmp = *mp;
						*mp = ep;
						ep->ent_next = tmp;
					}
					break;
				}
				vp = vp->avseq_next;
			}
		}

		ep = savenext;
	}
	if ( debug & 8 ) printf("returning (%d) matches\n", matches);
	return(matches);
}

as_find_value_type(list, key)
Attr_Sequence	list;
AV_Sequence	key;
{
	AV_Sequence	v;

	for ( v = list->attr_value; v != NULLAV; v = v->avseq_next )
		if ( AttrV_cmp( &v->avseq_av, &key->avseq_av ) == 0 )
			return( 0 );	/* found it */
	return( 1 );
}

as_find_value(list, key)
Attr_Sequence	list;
Attr_Sequence	key;
{
	Attr_Sequence	tas;
	AV_Sequence	tav;
	int		first;

	if (tas = as_find_type(list, key->attr_type)) {
		first = 1;
		for (tav = tas->attr_value; tav; tav = tav->avseq_next) {
			if (AttrV_cmp(&tav->avseq_av,
			    &key->attr_value->avseq_av) == 0)
				return((first && tav->avseq_next == NULLAV) ?
				    -1 : 0);
			first = 0;
		}
		return(2);	/* value not found */
	}

	return(1);	/* type not found */
}

AttributeValue rdntype(dn, type)
DN		dn;
AttributeType	type;
{
	while (dn->dn_parent)
		dn = dn->dn_parent;

	if (AttrT_cmp(type, dn->dn_rdn->rdn_at) == 0)
		return(&dn->dn_rdn->rdn_av);
	else
		return(NULLAttrV);
}

print_entrymod(em)
struct entrymod	*em;
{
	struct entrymod	*ep;

	printf("*** Entry Mod ***\n");
	while (em) {
		printf("em->em_what = ");
		fflush(stdout);
/*
		print_dn(em->em_what);
*/
		printf("em->em_type = ");
		switch (em->em_type) {
		case EM_ADDATTRIBUTE:
			printf("EM_ADDATTRIBUTE");
			break;
		case EM_ADDVALUES:
			printf("EM_ADDVALUES");
			break;
		case EM_REMOVEATTRIBUTE:
			printf("EM_REMOVEATTRIBUTE");
			break;
		case EM_REMOVEVALUES:
			printf("EM_REMOVEVALUES");
			break;
		default:
			printf("default");
			break;
		}

		for (ep = em; ep; ep = ep->em_next) {
		}
	}
}

chase_referral( re )
struct DSError	*re;
{
	ContinuationRef		cr;
	struct access_point	*ap;
	char			bindpacket[1024], replypacket[5];
	char			*save;
	int			wasbound;
	extern char		*dsa_address;
	extern char		*bound_dnstr;
	extern char		*bound_pw;
	extern int		bound;

	/*
	 * we got a referral.  chase it and bind to the new DSA.
	 */

	save = strdup( dsa_address );
	wasbound = bound;

	/* for each dsa candidate */
	for ( cr = re->ERR_REFERRAL.DSE_ref_candidates;
	    cr != NULLCONTINUATIONREF; cr = cr->cr_next ) {

		/* for each access point to the dsa */
		for ( ap = cr->cr_accesspoints; ap != NULLACCESSPOINT;
		    ap = ap->ap_next ) {
			free( dsa_address );
			dsa_address = strdup( paddr2str( ap->ap_address,
			    NULLNA ) );

			if ( debug )
				printf( "referring to (%s)\n", dsa_address );
			sprintf( bindpacket, "%s%c%s", bound_dnstr, 0,
			    bound_pw );
			if ( debug & 4 ) {
				printf( "calling bbindop with\n" );
				bprint( bindpacket, strlen( bound_dnstr ) + 
				    strlen( bound_pw ) + 2 );
			}
			bbindop( bindpacket, replypacket, 0 );

			if ( bound )
				break;
		}

		if ( bound )
			break;
	}

	if ( bound ) {
		free( save );
		return( OK );
	} else {
		free( dsa_address );
		dsa_address = save;
		if ( wasbound ) {
			sprintf( bindpacket, "%s%c%s", bound_dnstr, 0,
			    bound_pw );
			bbindop( bindpacket, replypacket, 0 );
		} else
			nullbind();
		return( NOTOK );
	}
}


clean_exit( rc )
{
    struct timeval	tv;

    if ( dosyslog ) {
	if ( stream_connected ) {
	    gettimeofday( &tv, (struct timezone *)NULL );
	    syslog( LOG_INFO|LOG_LOCAL1, "TCP closed %d seconds",
		    tv.tv_sec - stream_conn_start_time );
	}
	syslog( LOG_INFO|LOG_LOCAL1, "exit(%d)", rc );
    }

    exit( rc );
}
